@@ -5,12 +5,13 @@
 #' @name compute_lp-forestry
 #' @title compute lp distances
 #' @rdname compute_lp-forestry
-#' @description return lp ditances of selected test observations.
+#' @description Return the L_p norm distances of selected test observations
+#'   relative to the training observations which the forest was trained on.
 #' @param object A `forestry` object.
-#' @param feature.new A data frame of testing predictors.
+#' @param newdata A data frame of test predictors.
 #' @param feature A string denoting the dimension for computing lp distances.
 #' @param p A positive real number determining the norm p-norm used.
-#' @return A vector lp distances.
+#' @return A vector of the lp distances.
 #' @examples
 #'
 #' # Set seed for reproductivity
@@ -27,17 +28,17 @@
 #'
 #' # Compute the l2 distances in the "Petal.Length" dimension
 #' distances_2 <- compute_lp(object = rf,
-#'                           feature.new = x_test,
+#'                           newdata = x_test,
 #'                           feature = "Petal.Length",
 #'                           p = 2)
 #' @export
-compute_lp <- function(object, feature.new, feature, p){
+compute_lp <- function(object, newdata, feature, p){
 
   # Checks and parsing:
   if (class(object) != "forestry") {
     stop("The object submitted is not a forestry random forest")
   }
-  feature.new <- as.data.frame(feature.new)
+  newdata <- as.data.frame(newdata)
   train_set <- slot(object, "processed_dta")$processed_x
 
   if (!(feature %in% colnames(train_set))) {
@@ -46,10 +47,10 @@ compute_lp <- function(object, feature.new, feature, p){
 
   # Compute distances
   y_weights <- predict(object = object,
-                       feature.new = feature.new,
+                       newdata = newdata,
                        aggregation = "weightMatrix")$weightMatrix
 
-  if (is.factor(feature.new[1, feature])) {
+  if (is.factor(newdata[1, feature])) {
     # Get categorical feature mapping
     mapping <- slot(object, "categoricalFeatureMapping")
 
@@ -58,31 +59,31 @@ compute_lp <- function(object, feature.new, feature, p){
     # * Run it on data set with several factor valued covariates
     # * Implement a test with a data set with two factor-valued column
     # * Try to use:
-    # processed_x <- preprocess_testing(feature.new,
+    # processed_x <- preprocess_testing(newdata,
     #                                   object@categoricalFeatureCols,
     #                                   object@categoricalFeatureMapping)
     factor_vals <- mapping[[1]][2][[1]]
     map <- function(x) {
       return(which(factor_vals == x)[1])
     }
-    feature.new[ ,feature] <- unlist(lapply(feature.new[,feature], map))
+    newdata[ ,feature] <- unlist(lapply(newdata[,feature], map))
 
-    diff_mat <- matrix(feature.new[,feature],
-                       nrow = nrow(feature.new),
+    diff_mat <- matrix(newdata[,feature],
+                       nrow = nrow(newdata),
                        ncol = nrow(train_set),
                        byrow = TRUE) !=
                 matrix(train_set[,feature],
-                       nrow = nrow(feature.new),
+                       nrow = nrow(newdata),
                        ncol = nrow(train_set),
                        byrow = FALSE)
     diff_mat[diff_mat] <- 1
   } else {
-    diff_mat <- matrix(feature.new[,feature],
-                       nrow = nrow(feature.new),
+    diff_mat <- matrix(newdata[,feature],
+                       nrow = nrow(newdata),
                        ncol = nrow(train_set),
                        byrow = TRUE) -
                 matrix(train_set[,feature],
-                       nrow = nrow(feature.new),
+                       nrow = nrow(newdata),
                        ncol = nrow(train_set),
                        byrow = FALSE)
   }
@@ -94,7 +95,7 @@ compute_lp <- function(object, feature.new, feature, p){
   distances <- apply(y_weights * diff_mat, 1, sum) ^ (1 / p)
 
   # Ensure that the Lp distances for a factor are between 0 and 1
-  if (is.factor(feature.new[1, feature])) {
+  if (is.factor(newdata[1, feature])) {
     distances[distances < 0] <- 0
     distances[distances > 1] <- 1
   }
